#include <torch/extension.h>
#include <vector>
#include <cmath>

// Device helper functions for activations
__device__ inline float device_sigmoid(float x) {
    return 1.0f / (1.0f + expf(-x));
}

__device__ inline float device_tanh(float x) {
    return tanhf(x);
}

// Modular device function for LSTM cell computation
__device__ inline void compute_lstm_cell(const float* __restrict__ gates,
                                          int gate_base,
                                          float prev_c,
                                          int hidden_size,
                                          float &h_out,
                                          float &c_out) {
    float i_gate = device_sigmoid(gates[gate_base]);
    float f_gate = device_sigmoid(gates[gate_base + hidden_size]);
    float g_gate = device_tanh(gates[gate_base + 2 * hidden_size]);
    float o_gate = device_sigmoid(gates[gate_base + 3 * hidden_size]);
    float c_new = f_gate * prev_c + i_gate * g_gate;
    float tanh_c_new = device_tanh(c_new); h_out = o_gate * tanh_c_new;
    c_out = c_new;
}

// CUDA kernel for element-wise LSTM computations
__global__ void lstm_elementwise_forward_modular(
    const float* __restrict__ gates,
    const float* __restrict__ prev_c,
    float* __restrict__ h,
    float* __restrict__ c,
    int batch_size,
    int hidden_size
) {
    int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx < batch_size * hidden_size) {
        int b = idx / hidden_size;
        int n = idx % hidden_size;
        int gate_base = b * hidden_size * 4 + n;

        float h_val, c_val;
        compute_lstm_cell(gates, gate_base, prev_c[idx], hidden_size, h_val, c_val);
        h[idx] = h_val;
        c[idx] = c_val;
    }
}

// Tiled linear forward kernel using shared memory reduction
// Grid dimensions: grid.x = batch index, grid.y = output index
__global__ void linear_forward_kernel_tiled(
    const float* __restrict__ input,
    const float* __restrict__ weight,
    const float* __restrict__ bias,
    float* __restrict__ output,
    int input_dim
) {
    int batch_idx = blockIdx.x;  // Each block row corresponds to a unique batch element
    int out_idx   = blockIdx.y;  // Each block column corresponds to a unique output neuron
    int tid = threadIdx.x;
    float sum = 0.0f;

    // Each thread computes partial dot product across the input dimension
    for (int k = tid; k < input_dim; k += blockDim.x) {
        sum += input[batch_idx * input_dim + k] * weight[out_idx * input_dim + k];
    }

    // Shared memory reduction
    extern __shared__ float shmem[];
    shmem[tid] = sum;
    __syncthreads();

    // Reduction in shared memory
    for (int s = blockDim.x / 2; s > 0; s >>= 1) {
        if (tid < s) {
            shmem[tid] += shmem[tid + s];
        }
        __syncthreads();
    }

    // Write final result to global memory
    if (tid == 0) {
        float res = shmem[0];
        if (bias != nullptr) {
            res += bias[out_idx];
        }
        // Using gridDim.y as the output dimension stride
        output[batch_idx * gridDim.y + out_idx] = res;
    }
}

// Function to compute LSTM forward pass using the modular kernel
torch::Tensor lstm_forward_cuda(
    torch::Tensor input,
    torch::Tensor w_ih,
    torch::Tensor w_hh,
    torch::Tensor b_ih,
    torch::Tensor b_hh,
    torch::Tensor h0,
    torch::Tensor c0
) {
    auto options = torch::TensorOptions().dtype(input.dtype()).device(input.device());
    int batch_size = input.size(0);
    int seq_len = input.size(1);
    int hidden_size = h0.size(1);

    torch::Tensor h = h0.clone();
    torch::Tensor c = c0.clone();

    std::vector<torch::Tensor> outputs;

    for (int t = 0; t < seq_len; ++t) {
        // Get input at timestep t
        torch::Tensor xt = input.select(1, t);

        // Compute gates: gates = x_t * W_ih^T + h_{t-1} * W_hh^T + b_ih + b_hh
        torch::Tensor gates = torch::addmm(b_ih, xt, w_ih.t());
        gates = torch::addmm(gates, h, w_hh.t());
        gates += b_hh;

        // Prepare device pointers
        const float* gates_ptr = gates.data_ptr<float>();
        const float* prev_c_ptr = c.data_ptr<float>();
        float* h_ptr = h.data_ptr<float>();
        float* c_ptr = c.data_ptr<float>();

        int total_elements = batch_size * hidden_size;
        int threads = 256;
        int blocks = (total_elements + threads - 1) / threads;

        // Launch CUDA kernel for LSTM cell computations
        lstm_elementwise_forward_modular<<<blocks, threads>>>(
            gates_ptr,
            prev_c_ptr,
            h_ptr,
            c_ptr,
            batch_size,
            hidden_size
        );

        // Collect output
        outputs.push_back(h.unsqueeze(1));
    }

    // Concatenate outputs
    torch::Tensor output = torch::cat(outputs, 1);

    return output;
}

// Function to compute linear forward pass using tiled reduction without atomic operations
torch::Tensor linear_forward_cuda(
    torch::Tensor input,
    torch::Tensor weight,
    torch::Tensor bias
) {
    int batch_size = input.size(0);
    int input_dim = input.size(1);
    int output_dim = weight.size(0);

    auto options = input.options();
    auto output = torch::empty({batch_size, output_dim}, options);

    // Launch one block per (batch, output) pair
    dim3 grid(batch_size, output_dim);
    int threads = 256;
    size_t shared_mem_size = threads * sizeof(float);

    linear_forward_kernel_tiled<<<grid, threads, shared_mem_size>>>(
        input.data_ptr<float>(),
        weight.data_ptr<float>(),
        bias.defined() ? bias.data_ptr<float>() : nullptr,
        output.data_ptr<float>(),
        input_dim
    );

    return output;
}

// Main forward function integrating LSTM and Linear layers
torch::Tensor forward(
    torch::Tensor x,
    std::vector<torch::Tensor> lstm_weights_ih,
    std::vector<torch::Tensor> lstm_weights_hh,
    std::vector<torch::Tensor> lstm_biases_ih,
    std::vector<torch::Tensor> lstm_biases_hh,
    torch::Tensor fc_weight,
    torch::Tensor fc_bias,
    torch::Tensor h0,
    torch::Tensor c0,
    bool is_training
) {
    // Ensure h0 and c0 are on the same device as x
    h0 = h0.to(x.device());
    c0 = c0.to(x.device());

    torch::Tensor out = x;
    int num_layers = lstm_weights_ih.size();

    for (int i = 0; i < num_layers; ++i) {
        auto w_ih = lstm_weights_ih[i].to(x.device());
        auto w_hh = lstm_weights_hh[i].to(x.device());
        auto b_ih = lstm_biases_ih[i].to(x.device());
        auto b_hh = lstm_biases_hh[i].to(x.device());

        // Prepare initial hidden and cell states for the current layer
        torch::Tensor h_i = h0.narrow(0, i, 1).squeeze(0);
        torch::Tensor c_i = c0.narrow(0, i, 1).squeeze(0);

        // Run LSTM layer
        out = lstm_forward_cuda(out, w_ih, w_hh, b_ih, b_hh, h_i, c_i);
    }

    // Take the last timestep
    out = out.select(1, -1);

    // Apply final linear layer using the tiled reduction kernel
    out = linear_forward_cuda(out, fc_weight, fc_bias);

    return out;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Forward pass with fused LSTM and tiled linear kernel (CUDA)");
}
